/*******************************************************************/
/*  ASSIGNMENT #2     QUESTION 4
    Tokenizer module contains the following functions as 
    described in the specifications.

    PROCEDURE open_token_stream(Filename: string): token_stream;
      Associates the sequence of tokens in the file to a 
      token_stream.

    PROCEDURE get_token(T: token_stream): token;
      Gets the next token in the token_stream.
 
    PROCEDURE compare_token( T1, T2: token): BOOLEAN;
      Returns SUCCEED if the two tokens are equal.
      Otherwise returns FAIL.
      Tokens are considered equal if they are both
        - the same keyword, or
        - the same special symbol
        - constants of the same type (may be different values), or
        - any identifier (may have different variable name)
        If either of the tokens is error, they are not equal.

    PROCEDURE print_token(T: token): BOOLEAN;
      Prints the token according the the format
        If token has two elements print        element1, "element2".
        If token has only one element print    element1.

    PROCEDURE is_eof_token(T: token): BOOLEAN;
        If T represents the "eof" token, returns SUCCEED.
        Otherwise returns FAIL.                                    
    
    PROCEDURE unget_error(T: character_stream);
        Handles errors in unget_char().  It is defined in stream.c.
/*******************************************************************/


           
#include "tokens.h"


           
token_stream open_token_stream(Filename)
string Filename;
{
  character_stream S;

  if ((S = open_character_stream(Filename)) == NULL)
     return NULL;
  else
     return S;
}




token get_token(T)
token_stream T;
{
  token t1;
  CHARACTER c; 
  char word[80];
  char errflag;

  char comment();       /* PROCESSES A COMMENT TOKEN */
  char *num_const();    /* PROCESSES A NUMERIC CONSTANT TOKEN */
  char *char_const();   /* PROCESSES A CHARACTER CONSTANT TOKEN */
  char *str_const();    /* PROCESSES A STRING CONSTANT TOKEN */
  char *kwd_assign();   /* PROCESSES AN ASSIGNMENT KEYWORD TOKEN */ 
  char *keywd_id();     /* PROCESSES A KEYWORD/IDENTIFIER TOKEN */
  char *err_const();    /* PROCESSES AN ERROR */
  BOOLEAN is_keywd();   /* DETERMINES IF A TOKEN IS A KEYWORD OR 
                           AN IDENTIFIER */

  t1.e1 = "";
  t1.e2 = "";
  if (is_end_of_character_stream(T))
    t1.e1 = "eof";
  else
    {
      c = get_char(T);
      while ((isspace(c)) && (! is_end_of_character_stream(T)))
           c = get_char(T);        /* ELIMINATE WHITESPACE */

      if (c == ';')                /* COMMENT TOKEN */
          c = comment(T);

      if (c == '#')                /* CHARACTER CONSTANT TOKEN */
         {
           t1.e1 = "character";
           t1.e2 = char_const(T);
         }
    
      else if (c == '"')           /* STRING CONSTANT TOKEN */
         {
           t1.e1 = "string";
           t1.e2 = str_const(T,c);
         }
         
      else if (isdigit(c))         /* NUMERIC CONSTANT TOKEN */
         {
           errflag = '0';
           t1.e2 = num_const(T,c,&errflag);
           if (errflag == '1')
              t1.e1 = "error";
           else
              t1.e1 = "numeric"; 
         }

      else if (c == '=')           /* ASSIGNMENT KEYWORD TOKEN */
         {
           errflag = '0';
           t1.e2 = kwd_assign(T,c,&errflag);
           if (errflag == '1')
              t1.e1 = "error";
           else
              t1.e1 = "keyword";
         }

      else if (c == '(')           /* LPAREN TOKEN */
         {
           t1.e1 = "lparen";
           t1.e2 = "";
         }

      else if (c == ')')           /* RPAREN TOKEN */
         {
           t1.e1 = "rparen";
            t1.e2 = "";
         }

      else if (c == '[')           /* LSQUARE TOKEN */
         {
           t1.e1  = "lsquare";
            t1.e2 = ""; 
         }

      else if (c == ']')           /* RSQUARE TOKEN */
         {
           t1.e1 = "rsquare";
           t1.e2 = "";
         }

      else if (c == '\'')          /* QUOTE TOKEN */
         {
           t1.e1 = "quote";
            t1.e2 = ""; 
         }

      else if (c == '`')           /* BQUOTE TOKEN */
         {
           t1.e1 = "bquote";
           t1.e2 = "";
         }

      else if (c == ',')           /* COMMA TOKEN */
         {
           t1.e1 = "comma";
            t1.e2 = "";
         }

      else if (is_end_of_character_stream(T))        /* EOF TOKEN */
         {
           t1.e1 = "eof";
           t1.e2 = ""; 
         }

      else if (isalpha(c))         /* KEYWORD OR IDENTIFIER TOKEN */
         {
           errflag = '0';
           t1.e2 = keywd_id(T,c,&errflag);
           if (errflag == '1')
              t1.e1 = "error";
           else
              {
                if (is_keywd(t1.e2))
                   t1.e1 = "keyword";
                else
                   t1.e1 = "identifier"; 
              }
         }

      else
         {
           t1.e1 = "error";
           t1.e2 = err_const(c);
         }
    }
  return t1;
}





BOOLEAN compare_token(t1,t2)
token t1;
token t2;
{

  if (((strcmp(t1.e1, "keyword")) == 0) && ((strcmp(t1.e1, t2.e1)) == 0) && ((strcmp(t1.e2, t2.e2)) == 0))
     return 1;
  else if (((strcmp(t1.e1, t2.e1)) == 0) && ((strcmp(t1.e1,"error")) != 0) && ((strcmp(t2.e1, "error")) != 0))
     return 1;
  else
     return 0;
}





BOOLEAN print_token(t)
token t;
{
  if (((strcmp(t.e2,"")) == 0) && ((strcmp(t.e1,"string")) != 0))
     printf("%s.\n",t.e1);
  else if ((strcmp(t.e1, "numeric")) == 0)
     printf("%s, %s.\n",t.e1,t.e2);
  else
     printf("%s, \"%s\".\n",t.e1,t.e2);
}





BOOLEAN is_eof_token(t)
token t;
{
  if ((strcmp(t.e1, "eof")) == 0)
     return 1;
  else
     return 0;
}




/**********************************************************************/
/*  AUXILIARY FUNCTIONS - not defined in the specifications
    Performs the functions as described in PROCEDURE get_token        */
/**********************************************************************/


char comment(T)
token_stream T;
{
  CHARACTER c;
  
  c = get_char(T);
  while ((c != '\n') && (! is_end_of_character_stream(T)))
       c = get_char(T);
      
  while (isspace(c))
       c = get_char(T);

  if (c == ';')
     comment(T);
  else
     return c;
}





char *num_const(T,c,eflag)
token_stream T;
CHARACTER c;
char *eflag;
{
  CHARACTER x;
  char num[80];
  int i;
  
  i = 0;
  num[i++] = c;
  x = get_char(T);
  while (! isspace(x))
       {
         if (! isdigit(x))
            {
              num[0] = x;
              num[1] = '\0';
              *eflag = '1';
              return num;
            }
         else if (*eflag != '1')
            {
              num[i++] = x;
              x = get_char(T);
            }
       }
  num[i] = '\0';
  return num;
} 





char *kwd_assign(T,c,eflag)
token_stream T;
CHARACTER c;
char *eflag;
{
  CHARACTER x;
  char kwda[80];
  
  kwda[0] = c;
  x = get_char(T);
  if (x == '>')
     {
       kwda[1] = x;
       kwda[2] = '\0';
     }
  else
     {
       unget_char(x,T);
       kwda[1] = '\0';
       *eflag = '1';
     }
  return kwda;
}





char *str_const(T,c)
token_stream T;
CHARACTER c;
{
  CHARACTER x;
  char str[80];
  int i;
  
  i = 0;
  x = get_char(T);
  if (is_end_of_character_stream(T))
     {
       printf("FATAL ERROR: unterminated character string.\n");
       exit(0);
     }
  while ((x != '"') && (! is_end_of_character_stream(T)))
     {
       str[i++] = x;
       x = get_char(T);
       if (is_end_of_character_stream(T))
          {
            printf("FATAL ERROR: unterminated character string.\n");
            exit(0);
          }
     }
  str[i] = '\0';
  return str;
}





char *keywd_id(T,c,eflag)
token_stream T;
CHARACTER c;
char *eflag;
{
  CHARACTER x;
  char kwd[80];
  int i;
  
  i = 0;
  kwd[i++] = c;
  x = get_char(T);
  while (! isspace(x))
     {
       if (! isalnum(x))
          {
            kwd[0] = x;
            kwd[1] = '\0';
            *eflag = '1';
            return kwd; 
          }
       else 
          {
            kwd[i++] = x;
            x = get_char(T);
          }
     }
  kwd[i] = '\0';
  return kwd;
} 





BOOLEAN is_keywd(s)
CHARACTER *s;
{
  if ((strcmp(s,"and")) == 0)
     return 1;
  else if ((strcmp(s,"or")) == 0) 
     return 1;
  else if ((strcmp(s,"if")) == 0) 
     return 1;
  else if ((strcmp(s,"xor")) == 0) 
     return 1;
  else if ((strcmp(s,"lambda")) == 0)
     return 1;
  else
     return 0;
}





char *char_const(T)
token_stream T;
{
  CHARACTER x;
  char ch_const[80];
  
  x = get_char(T);
  ch_const[0] = x;
  ch_const[1] = '\0';
  return ch_const;
}





char *err_const(c)
CHARACTER c;
{
  char e_const[80];
  
  e_const[0] = c;
  e_const[1] = '\0';
  return e_const;
}





void unget_error(T)
character_stream;
{
  printf("Unget error.\n");
  exit(0);
}
