/*  PROGRAM NAME: Compare.c 
 *  INPUTS      : filename1, filename2
 *  OUTPUTS     : result of comparison 
 *  DESCRIPTION : This program will accept two filenames from the command
 *		  line.  These two files contain scheme code.  If the
 *		  two files are equal or not equal, the result is given    
 *		  in standard output.
 *		  Equality is determined by the following criteria
 * 			1.  same keyword
 *			2.  same special symbol
 * 			3.  constants of same type (may have diff. value)
 * 			4.  any identifier (may have diff. name)
 */  


#include  <stdio.h> 
#include  "tokens.h" 

void main(argc,argv)
int argc;
char *argv[];
{
   TOKEN_STREAM stream1,stream2;
   TOKEN token1,token2;
   int status;

   /* If two arguments are not given for the filenames, the program
    * is terminated.  Try again.
    */ 
   if (argc != 3)
      {
	printf("You must enter two filenames\n"); 
       	exit();
      }

   /* open streams for both input files */
   stream1 = open_token_stream(argv[1]);
   stream2 = open_token_stream(argv[2]);

   token1 = get_token(&stream1);
   token2 = get_token(&stream2);
   status = compare_token(token1,token2);

   /* while the two files compare, keep getting the next token from
    * each file.  If they are not equal a status is given.
    */
   while ((!status) && (!is_eof_token(token1))
				 && (!is_eof_token(token2)))
      {
   	token1 = get_token(&stream1);
   	token2 = get_token(&stream2);
        status = compare_token(token1,token2);
      }

   if (!status)
   	printf("The two files are equal\n");
   else
   	printf("The two files are not equal\n");
} 

