/*  MODULE     : stream.c
 *  DESCRIPTION: This module performs various functions on a
 *  		 character stream.
 */
   
#include "stream.h"
#include <stdio.h>

/************************************************************/
/*  FUNCTION   : open_character_stream
 *  INPUTS     : filename
 *  OUTPUTS    : CHARACTER_STREAM
 *  DESCRIPTION: reads a stream of characters from a file of stdin 
 *               and associates it to character stream.
 */

CHARACTER_STREAM open_character_stream(filename) 
char *filename; 
{
   FILE *infile;
   CHARACTER_STREAM stream;

   stream.pos = 0;
   if (filename != NULL)
      {
      	if ((infile = fopen(filename,"r"))==NULL)
	      printf("Unable to open file specified\n");
   	else  
           {
 	      fgets(stream.string,sizeof(stream.string),infile);  
           }
      }
   else  /* get input from keyboard */ 
      {
	printf("Enter in scheme code to be processed\n");
	fgets(stream.string,sizeof(stream.string),stdin);  
      }
   return(stream);   
}

/************************************************************/
/*  FUNCTION   : get_char 
 *  INPUTS     : character stream 
 *  OUTPUTS    : char 
 *  DESCRIPTION: get a character from character stream.  
 */

char get_char(stream)  
CHARACTER_STREAM *stream;
{
   return(stream->string[stream->pos++]); 
}
 
/************************************************************/
/*  FUNCTION   : is_end_of_character_stream 
 *  INPUTS     : character stream 
 *  OUTPUTS    : boolean 
 *  DESCRIPTION: if the end of the character stream is reached
 *  		 then TRUE is returned, else FALSE is returned  
 */

is_end_of_character_stream(stream) 
CHARACTER_STREAM stream;
{
   if (stream.string[stream.pos] == NULL) 
	return(TRUE);
   else 
	return(FALSE);
}

/************************************************************/
/*  FUNCTION   : push_char 
 *  INPUTS     : character, character stream 
 *  OUTPUTS    : none 
 *  DESCRIPTION: this function "puts a character back on the
 * 		 stream" by decrementing the position index.
 *	 	 if there is no character to push, then error.  
 */

void push_char(c,stream)
char c;
CHARACTER_STREAM *stream;
{
   if (stream->pos == 0) 
      unget_error(stream);
   else
      {
	--stream->pos;
	stream->string[stream->pos] = c; 
      }
}

/************************************************************/
/*  FUNCTION   : unget_error 
 *  INPUTS     : character stream 
 *  OUTPUTS    : none 
 *  DESCRIPTION: called if there is no character to be pushed 
 *  		 back on the character stream. 
 */

void unget_error(stream)
CHARACTER_STREAM stream;
{
   printf("Unable to unget character\n");
}
