/*  HEADER FILE: tokens.h 
 *  DESCRIPTION: This file defines all structures and types needed to
 *   		 represent TOKENs and TOKEN_STREAMs in tokenizer module.  
 *		 All functions used by token module are also defined
 * 		 below
 */      

/*  ctype.h is an include file that contains functions to determine if
 *  a particular character is alphabetic, numeric, or a whitespace.
 */
#include <ctype.h>

#define FALSE 0
#define TRUE  1

/* This macro determines if a character is a valid delimiter */
#define symbol(c)(((c=='(')||(c==')')||(c=='[')||(c==']')||(c=='\'')||(c=='`')||(c==',')||(c=='#')||(c=='=')) ? 1 : 0)  

#define MAX_NUMBER_OF_TOKENS 256/* No. of tokens possible in stream */ 
#define DESCRIPTOR_LENGTH    11 /* possible length of token names */
#define MAX_TOKEN_SIZE       8  /* possible length for strings */

/*  The following declarations make up the user-defined types of 
 *  TOKEN and TOKEN_STREAM.
 *  A token consists of two elements.  The first element represents what
 *  the type of the token is.  The second element represents what the 
 *  actual token read is.
 *  A token stream consist of an array of type tokens.  The position
 *  of the stream is kept track by an index.
 */

struct token_record {
	char first_element[DESCRIPTOR_LENGTH];
	char second_element[MAX_TOKEN_SIZE]; 
	};	 
typedef struct token_record TOKEN; 

struct token_stream_record 
   {
	TOKEN token_array[MAX_TOKEN_SIZE];
	int  pos; 
   }; 
typedef struct token_stream_record TOKEN_STREAM; 

TOKEN_STREAM open_token_stream();
void	     remove_comments();
TOKEN 	     get_symbol_keyword(); 
TOKEN 	     get_alpha_numeric();
int	     is_keyword();
TOKEN 	     get_string_constant();
TOKEN 	     get_numeric_constant();
TOKEN 	     get_character_constant();
TOKEN 	     error_token();
TOKEN        get_string();
int	     remove_whitespace();
TOKEN        make_token();
TOKEN	     get_token();
int          compare_token();
int          print_token();
int          is_eof_token();


