/***********************************************************/
/*    tokenizer module             */
/***********************************************************/


/*************************************************************
* File name   : tokenizer.c
* Purpose     : To accept a token stream and perform various
*               functions like creating a token stream, retur-
*               ning a token,etc ...
* Needs       : Header files stream.h and tokens.h
* Conventions : Variables of type token_stream start with le-
*               tters Token
* External fns.: character stream open_character_stream(filen-
*                   ame : string);
*                char get_char(stream :character_stream);
*************************************************************/
#include <stdio.h>
#include "stream.h"              
#include "tokens.h"
#define MAX_TOKEN_LENGTH 100
#define NO_KEYWORDS 6
#define ERROR 12
#define MAX 100
#define END_TOKEN_CODE 12 


/***********************************************************
** Module name     : init(arg1, arg2)
** inputs          : a character array & an integer
** outputs         : none
** purpose         :initialises array to null
** funs. called    : Nil.
***********************************************************/

void init(buffer,i)
 char *buffer;
 int i;
  {
   int j;
   j = i;
   while(j >= 0){
     buffer[j] = '\0';
     --j;
    }
  }


/***********************************************************
** Module name     : compare_token(arg1,arg2)
** inputs          : Two tokens
** outputs         : Boolean value
** purpose         : Compares two tokens based on their code
**                   and flags 1 or 0
** fns. called     : Nil
***********************************************************/

BOOLEAN compare_token(Token1,Token2)
 token Token1,Token2;
{
  int value1,value2;
  value1 = Token1->tkn_code;
  value2 = Token2->tkn_code;
  if(value1 == value2)return 1;
  else return 0;
}

/***********************************************************
** Module name   : make_token(arg1,arg2,arg3)
** inputs        : Array of chars.,an integer which correspo-
**                 nds to the code of token and a token_stre-
**                 am pointer 
** outputs       : A token_stream pointer
** purpose       : Creates a linked list of tokens
** fns. called   : (token_stream)malloc(arg:int)    
************************************************************/

token_stream make_token(buffer, code,Token)
char buffer[];
int code;
token_stream Token;
{
  token_stream temp,track;
    
    track = Token;
   temp = (token_stream) malloc(sizeof(struct token));
   temp->message = (char *)malloc( (unsigned)(strlen( buffer) + 1));
   strcpy(temp->message,buffer);
   temp->tkn_code = code;
   temp->next = (token_stream)0;
   if(track == (token_stream)0){
      track = temp;
      Token = temp;
     }
   else{
        while(track->next != (token_stream) 0) track = track->next;
      track->next = temp;
     }
     return(Token);

}

/*************************************************************
** Module name    : check(arg1)
** inputs         : A character
** outputs        : An integer which corresponds to the state
**                  of the Finite Automata
** purpose        : Checks the state to which the automata tr-
**                  ansforms to.
** Functions      :(int)isalpha(ch:char);
**                  (int)isdigit(ch:char); 
**************************************************************/ 
 
int check(ch)
 char ch;
{
  int state;
  switch(ch){
        case ';' : state = 9;
                   break;
        case '(' : 
                  
        case ')' :
                 
        case '[' : 
                    
        case ']' :
                  
        case '\'':
                   
        case '\`':
                      
        case ',': state = 11;
                   break;     
        case ' ' : state = 21;
                   break;     
        case '\n': state = 23;
                   break;     
        case '#' : state = 7;
                   break;
        case '\"': state = 5;
                   break;
        case EOF : state = 24;
                   break;
        case '=' : state = 26;
                   break;
        default  :
                  if(isalpha(ch))state = 1;
                  else if(isdigit(ch)) state = 3;
                       else state = 25;
                   break;     
        }
    return(state);
}/*end of fn. check(arg). */


/***********************************************************
** Module name   : open_token_stream(arg1)
** inputs        : A file name 
** outputs       : A token_stream pointer
** purpose       : Returns a pointer to alist of tokens.
** Functions     : (int)check(ch:char);
**                 (void)init(arg1:array of char,arg2:int);
**                 (token_stream)make_token(arg1,arg2,arg3)
**                    arg1 : array of char;
**                    arg2 : int;
**                    arg3 : token_stream;
**                 (int)isalpha(ch:char);
**                 (int)isdigit(ch:char);
**                 (char)get_char(stream : character_stream);
**                 (character_stream)open_character_stream(
**                    stream : character_stream);
**                 (int)is_end_of_character_stream(stream)
**                        stream :character_stream;
**                 (void)unget_error(ch : char;stream :chara-
**                       cter_stream;
***********************************************************/

token_stream open_token_stream(filename)
 string filename;
  {
   
   character_stream stream;
   token_stream token_head;
   CHARACTER ch, buffer[MAX];
   int state,code,flag,lflag,i = -1;
   token_head = (token_stream )0;
   stream = (character_stream)open_character_stream(filename);
    ch = '(';
       ch = get_char(stream);
        state = check(ch);
        flag = 1;
  do{
     switch(state){
        case 1: 
               
               while(flag)
                if(isalpha(ch) || isdigit(ch)){
                  buffer[++i] = ch;
                  if(!is_end_of_character_stream(stream))
                   ch = get_char(stream);
                  else { 
                   state = 24;
                   flag = 0;
                   }

                 }
                else{
                 state = 2;
                 flag = 0;
                 }
               break;
        case 2:
               code = 1;
               unget_char(ch,stream);
               buffer[++i] = '\0';
               token_head = make_token(buffer,code,token_head);
                  init(buffer,i);
                   i = -1;
              
                if(!is_end_of_character_stream(stream)){
                  ch = get_char(stream);
                  state = check(ch);
                  flag = 1;
                  }
                else state = 24;
                  break;
           case 3:
                  while(flag)
                    if(isdigit(ch)){
                        buffer[++i] = ch;
                      if(!is_end_of_character_stream(stream))
                        ch = get_char(stream);
                      else { state = 24;
                          flag = 0;
                           }
                     }
                    else{
                       state = 4;
                       flag = 0;
                     }
                    break;
            case 4:
                   code = 2;
                   buffer[++i] = '\0';
                  unget_char(ch,stream);
       token_head = (token_stream)make_token(buffer,code,token_head);
                 init(buffer,i);
                   i = -1;
                if(!is_end_of_character_stream(stream)){
                  ch = get_char(stream);
                  state = check(ch);
                  flag = 1;
                  }
                else state = 24;
                  break;
            case 5:
                          buffer[++i] = ch;
                    lflag = 1;
                    while(lflag)
                       if(!is_end_of_character_stream(stream)){
                          ch = get_char(stream);
                          if(ch == '"'){
                              lflag = 0;
                              state = 6;
                           }
                          buffer[++i] = ch;
                         }
                       else{
                               i = -1;
                              buffer[++i] = '\"';
                              buffer[++i] = '\0';
                               code = 0;	
           token_head = (token_stream)make_token(buffer,code,token_head);
                            state = 24;
                            flag = 0;
                           }
                    break;
            case 6:
                    code = 3;
      token_head = (token_stream)make_token(buffer,code,token_head);
                    init(buffer,i);
                    i = -1;
                    if(!is_end_of_character_stream(stream)){
                     ch = get_char(stream);
                     state = check(ch);
                     flag = 1;
                     }
                    else state = 24;
                    break;
            case 7:
                    code = 4;
                  ch = get_char(stream);
                   buffer[0] = ch;
                   buffer[1] = '\0';
        token_head = (token_stream)make_token(buffer,code,token_head);
                    init(buffer,i);
                    i = -1;
                    if(!is_end_of_character_stream(stream)){
                     ch = get_char(stream);
                     state = check(ch);
                     flag = 1;
                     }
                    else state = 24;
                    break;
            case 9:
                    lflag = 1;
                    while(lflag)
                       if(!is_end_of_character_stream(stream)){
                          ch = get_char(stream);
                          if(ch == ';'){
                              lflag = 0;
                          if(!is_end_of_character_stream(stream)){
                                  ch = get_char(stream);
                                   state = check(ch);
                                   flag = 1;
                                  }
                            }
                          else state = 24;
                        }
                       else{
                            code = 0;
                            i = -1;	
                           buffer[++i] = ';';
                           buffer[++i] = '\0';
           token_head = (token_stream)make_token(buffer,code,token_head);
                           state = 24;
                            lflag = 0;
                           }
                     break;
                    
            case 11:
                    switch(ch){
                        case '(' : code = 5; 
                                   break;
                        case ')' : code = 6;
                                   break;
                        case '[' : code = 7;
                                   break;
                        case ']' : code = 8;
                                   break;
                        case '\'' : code = 9;
                                   break;
                        case '\`' : code = 10;
                                   break;
                        case ',' : code = 11;
                                   break;
                      }
                    buffer[i] = ch;
             token_head = (token_stream)make_token(buffer,code,token_head);
                   init(buffer,i);
                    i = -1;
                    if(!is_end_of_character_stream(stream)){
                     ch = get_char(stream);
                     state = check(ch);
                     flag = 1;
                     }
                    else {
                          state = 24;
                          }
                    break;
                    
            case  21:
                     while(ch == ' ')
                    if(!is_end_of_character_stream(stream)){
                     ch = get_char(stream);
                     state = check(ch);
                     flag = 1;
                     }
                    else state = 24;
                    break;
            case 23:
                    if(!is_end_of_character_stream(stream)){
                     ch = get_char(stream);
                     state = check(ch);
                     flag = 1;
                     }
                    else state = 24;
                    break;
            case 24:
                    code = 12;
                    buffer[1] = 'e';
                    buffer[2] = 'o';
                    buffer[3] = 'f';
                    buffer[4] = '\0';
        token_head = (token_stream)make_token(buffer,code,token_head);
                    state = 64;
                    break;
            case 25:
                    code = 0;
                    buffer[0] = ch;
                    buffer[1] = '\0';
      token_head = (token_stream)make_token(buffer,code,token_head);
                    if(!is_end_of_character_stream(stream)){
                     ch = get_char(stream);
                     state = check(ch);
                     flag = 1;
                     }
                    else state = 24;
                    break;
            case 26:
                      buffer[++i] = ch;
                    if(!is_end_of_character_stream(stream)){
                     ch = get_char(stream);
                     if(ch == '>'){code = 23;
                      buffer[++i] = ch;
                      }
                     else {
                           code = 0;
                           unget_char(ch,stream);
                         }
                      buffer[++i] = '\0';
      token_head = (token_stream)make_token(buffer,code,token_head);
                   init(buffer,i);
                    i = -1;
                    if(!is_end_of_character_stream(stream)){
                     ch = get_char(stream);
                     state = check(ch);
                     flag = 1;
                     }
                    else state = 24;
                    break;
                    }
            }

         }while(state != 64);/* end of while */
     return(token_head);

}



/***********************************************************
** Module name     : get_token(arg1,arg2)
** inputs          : Token_stream
** outputs         : token        
** purpose         : returns a token from the token stream. 
** fns. called     : Nil
***********************************************************/

token get_token(Token)
struct token *Token;
{
  token temp,to_ken;
  temp = (struct token *)malloc(sizeof(struct token));
  temp->message = Token->message;;
  temp->next = Token->next;
  temp->tkn_code = Token->tkn_code;
   to_ken = Token->next;
   while(to_ken != NULL){
      Token->message = to_ken->message;
      Token->tkn_code = to_ken->tkn_code;
      Token->next = to_ken->next;
      to_ken = to_ken->next;
  Token = Token->next;
   }
  return(temp);
}



/***********************************************************
** Module name     : is_eof_token(Token : token)  
** inputs          : token
** outputs         : boolean      
** purpose         : flags true if the token is eof.        
** Functions       : Nil
***********************************************************/
BOOLEAN is_eof_token(Token)
token Token;
{
 int value;
  if(Token->tkn_code == END_TOKEN_CODE) value = 1;
  else value = 0;
 return(value);
}


/***********************************************************
** Module name     : key_word(arg : token)  
** inputs          : token
** outputs         : boolean      
** purpose         : flags true if the token is a key word  
** Functions       : strcmp(str1,str2)
**                   str1,str2 : array of char;
***********************************************************/

BOOLEAN key_word(Keyword)
token Keyword;
{
  static char keys[5][10] = 
                           { "and","or","if","xor","lambda"};
    int i,flag=0;
    for(i=0;i < NO_KEYWORDS; i++)
           if(!(strcmp(keys[i],Keyword->message))){
            flag = 1;
           }
    return(flag);   
}



/***********************************************************
** Module name     : print_token(arg : token)
** inputs          : token
** outputs         : boolean      
** purpose         : flags true if the token is  printed
** Functions       : key_word(arg1 : token) 
**                   printf(arg)
**                   arg : pointer to an array
***********************************************************/

BOOLEAN print_token(Token)
token Token;
{
 int value=1;
 switch (Token->tkn_code){
     case 0:
            printf("error,      \"%s\". \n",Token->message);
            break;
     case 1:
      if(!key_word(Token))printf("identifier, \"%s\"\n",Token->message);
            else printf("keyword \" %s \". \n",Token->message);
            break;
     case 2:
            printf("numeric,    %s. \n",Token->message);
            break;
     case 3:
            printf("string,     %s. \n",Token->message);
            break;
     case 4:
            printf("character.  \"%s\". \n",Token->message);
            break;
     case 5:
            printf("lparen.\n");
            break;
     case 6:
            printf("rparen.\n");
            break;
     case 7:
            printf("lsquare.\n");
            break;
     case 8:
            printf("rsquare.\n");
            break;
     case 9:
            printf("quote.\n");
            break;
     case 10:
            printf("bquote.\n");
            break;
     case 11:
            printf("comma.\n");
            break;
     case 12:
            printf("eof.\n");
            break;
     case 23:
            printf("keyword,    \"%s\".\n",Token->message);
             break;
     }
  return(value);
}    
 
           
              
