#include <stdio.h>

#include "line_storage.h"

#define BUFFER_SIZE 128

/****************************************************************************
Function:	usage
Parameters:	none
Returns:	void
Purpose:	This function prints out the correct command-line usage of kwic
****************************************************************************/

void usage ()
{
  fprintf (stderr, "\nUsage:  kwic <filename>\n\n");
  exit (0);
}


/****************************************************************************
Function:	error
Parameters:	one "C type" string
Returns:	void
Purpose:	This function prints out str to stderr, if not NULL 
		and then exits the program.
****************************************************************************/

void error (char *str)
{
  if (str != NULL)
    fprintf (stderr, "\n%s\n\n", str);
  exit (1);
}


 
/****************************************************************************
Function:	fgetstr
Parameters:	FILE pointer to previously opened file
Returns:	a "C type" string
Purpose:	This function reads a string of arbitrary length from the 
		previously opened file, datafile, and returns that
		string.
****************************************************************************/

char *fgetstr (FILE *datafile)
{
  int count, curr_char, size;
  char *string;
  
  if (datafile == NULL)
    error ("kwic, fgetstr:  Bad parameter");
  size = BUFFER_SIZE;
  string = (char *) malloc (BUFFER_SIZE * sizeof (char));
  if (string == NULL)
    error ("kwic, fgetstr:  Memory allocation error");
  count = -1;
  do
  {
    count++;
    curr_char = getc (datafile);
    if ((curr_char == EOF) && (count == 0))
    {
      free (string);
      return (NULL);
    }
    string [count] = (char) curr_char;
    if ((count + 1) == size)
    {
      size = size + BUFFER_SIZE;
      string = (char *) realloc (string, size * sizeof (char));
      if (string == NULL)
	error ("kwic, fgetstr:  Memory allocation error");
    }
  } while ((string [count] != '\n') && (curr_char != EOF));
  string [count] = '\0';
  return (string);
}


 
/****************************************************************************
Function:	process_lines
Parameters:	FILE pointer to previously opened file
Returns:	void
Purpose:	This function does the main processing loop in kwic.  It
		reads in strings from datafile, generates the kwic index using 
		line_storage and then prints out the index.
****************************************************************************/

void process_lines (FILE *datafile)
{
  char *str;
  int count, num_lines;
  line_storage *storage;
  
  if (datafile == NULL)
    error ("kwic, process_lines:  Bad parameter");
  storage = empty_line_storage ();
  str = fgetstr (datafile);
  while (str != NULL)
  {
    circulate_and_add_line (storage, str);
    free (str);
    str = fgetstr (datafile);
  }
  storage = alphabetize (storage);
  num_lines = num_of_lines (storage);
  for (count = 1; count <= num_lines; count++)
  {
    str = get_line (storage, count);
    if (str == NULL)
      error ("kwic, process_lines:  Error getting line");
    printf ("%s\n", str);
    free (str);
  }
}


 
/****************************************************************************
Function:	main
Parameters:	argc, argv[]
Returns:	void
Purpose:	Main processes the command-line parameters and calls
		the appropriate functions.
****************************************************************************/

void main (int argc, char *argv[])
{
  FILE *datafile;
  
  switch (argc)
  {
    case 2:
             datafile = fopen (argv[1], "r");
	     if (datafile == NULL)
	       error ("kwic:  Error opening file");
	     process_lines (datafile);
                     fclose (datafile);
	     break;
    default:
	     usage ();
	     break;
  }
}  
