/****************************************************************************
Module:	line_storage

This module implements a type called line_storage and some operations that
can be performed on this type.  The currently supported operations are:

	empty_line_storage	-	creates empty line storage
	circulate_and_add_line	-	adds line and all circular shifts to
					line storage
	get_line			-	gets an arbitrary line from line storage
	num_of_lines		-	counts the lines in line storage
	alphabetize		-	sorts the lines in alphabetical order
****************************************************************************/

#ifndef _LINE_STORAGE_H
#define _LINE_STORAGE_H

#include <stdio.h>

#include "line.h"


/* Typedefs */

typedef struct _node
{
  line *contents;
  struct _node *next;
} node;

typedef struct
{
  node *head;
  int num_of_lines;
} line_storage;


/****************************************************************************
Function:	empty_line_storage
Parameters:	none
Returns:	a pointer to type line_storage
Purpose:	This function creates and returns a variable of type line_storage.
****************************************************************************/

line_storage *empty_line_storage ();


 
/****************************************************************************
Function:	circulate_and_add_line
Parameters:	one pointer to type line_storage and one "C type" string
Returns:	a pointer to type line_storage
Purpose:	This function takes string, converts it to a line and then
		adds all of the circular shifts of that line to storage.
Note:		Variable(s) passed in may be modified after function call.
****************************************************************************/

line_storage *circulate_and_add_line (line_storage *storage, char *string);


/****************************************************************************
Function:	get_line
Parameters:	one pointer to type line_storage and one integer
Returns:	a "C type" string
Purpose:	This function returns the line specified by position in 
		storage as a string.  Program will exit if position is
		out of range.
****************************************************************************/

char *get_line (line_storage *storage, int pos);


/****************************************************************************
Function:	num_of_lines
Parameters:	one pointer to type line_storage
Returns:	an integer
Purpose:	This function returns the number of lines in storage.
****************************************************************************/

int num_of_lines (line_storage *storage);


/****************************************************************************
Function:	alphabetize
Parameters:	one pointer to type line_storage
Returns:	a pointer to type line_storage
Purpose:	This function if supposed to sort the lines in storage
		alphabetically, but in this implementation the lines are
		always sorted.
****************************************************************************/

line_storage *alphabetize (line_storage *storage);


 
/****************************************************************************
Function:	dump_line_storage
Parameters:	one "C type" string and one pointer to type line_storage
Returns:	void
Purpose:	This function prints message to stderr and then proceeds to
		"dump" the entire contents of storage.
****************************************************************************/

void dump_line_storage (char *message, line_storage *storage);



#endif

